/*   
 * $Id: calc.c,v 1.21 2002/10/08 21:26:30 ondrap Exp $
 * 
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 1999,2000,2001 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
 * 
 *  Antonio Fiol Bonnn" <Antonio.FIOL@enst-bretagne.fr> submitted patches 
 *   for the financial part.
*/

#include <PalmOS.h>
#include <StringMgr.h>
#include <stdio.h>

#include "defuns.h"
#include "calc.h"
#include "calcDB.h"
#include "result.h"
#include "defmgr.h"
#include "calcrsc.h"    
#include "graph.h"
#include "finance.h"
#include "varmgr.h"
#include "prefs.h"
#include "memo.h"
#include "main.h"
#include "history.h"
#include "fp.h"
#include "lstedit.h"
#include "mtxedit.h"

#ifdef SPECFUN_ENABLED
#include "solver.h"
#endif

#ifdef GRAPHS_ENABLED
 
#include "grprefs.h"
#include "grsetup.h"
#include "graph.h"
#include "grtable.h"

#endif /* GRAPH_ENABLED */

Boolean palmOS35=false;
Boolean colorDisplay=false;
Boolean palmOS31=false;

void *GetObjectPtr(UInt16 objectID)
{
	FormPtr frm;
	
	frm=FrmGetActiveForm();
	return (FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,objectID)));
}

void
gadget_bounds(Int16 gadget,RectangleType *bounds)
{
	FormPtr frm = FrmGetActiveForm();
	UInt16 index = FrmGetObjectIndex(frm,gadget);
	
	FrmGetObjectBounds(frm,index,bounds);
}

static void 
calc_init_application()
{
	prefs_read_preferences();
}

/* Change to the next screen */
void
GotHardKey(void)
{
	int formId;
	
	switch (FrmGetActiveFormID()) {
	case frmBasic:
	case frmBasicS:
	    formId = frmScientific;
	    break;
	case frmScientific:
	    formId = frmInteger;
	    break;
	case frmInteger:
#ifdef GRAPHS_ENABLED
	    formId = frmGraph;
#else
	    if (calcPrefs.dispScien)
		formId = frmBasicS;
	    else
		formId = frmBasic;
#endif
	    break;
	default:
	case frmGraph:
	    if (calcPrefs.dispScien)
		formId = frmBasicS;
	    else
		formId = frmBasic;
	    break;
	}
	ChangeForm(formId);
}

/* Generic button action - actions for buttons  on all screens */
Int16
chooseForm(Int16 button)
{
	Boolean handled = true;

	switch (button) {
	case hlpMain2:
		FrmHelp(hlpMain2);
		break;
	case hlpFunc:
		FrmHelp(hlpFunc);
		break;
	case mitCredits:
		FrmHelp(strCredits);
		break;
	case mitAbout:
		FrmAlert(altAbout);
		break;
	case btnInteger:
		ChangeForm(frmInteger);
		break;
	case btnBasic:
		if (calcPrefs.dispScien) 
			ChangeForm(frmBasicS);
		else
			ChangeForm(frmBasic);
		break;
	case btnScientific:
		ChangeForm(frmScientific);
		break;
#ifdef GRAPHS_ENABLED		
	case btnGraph:
		ChangeForm(frmGraph);
		break;
#endif		
	default:
		handled = false;
	}
	return handled;
}

Boolean
calc_rom_greater(UInt16 major, UInt16 minor)
{
	UInt32        romVersion;
	UInt16        maj;
	UInt16        min;

	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	maj = sysGetROMVerMajor(romVersion);
	min = sysGetROMVerMinor(romVersion);
	if (maj > major)
		return true;
	else if ((maj == major) && (min >= minor))
		return true;
	
	return false;
}

static Err 
StartApplication(void)
{
	Err error;

	/* Get version of PalmOS */
	palmOS35 = calc_rom_greater(3,5);
	palmOS31 = calc_rom_greater(3,1);
	
	if (palmOS35)
	  WinScreenMode(winScreenModeGet,NULL,NULL,NULL,&colorDisplay);
	
	/* Initialize all modules */
	mlib_init_mathlib();
	
	error = db_open();
	ErrFatalDisplayIf(error, "Can't open CalcDB");
	
	error = history_open();
	ErrFatalDisplayIf(error, "Can't open History DB");
	
	calc_init_application();
	
	fp_setup_flpoint();

#ifdef GRAPHS_ENABLED	
	grpref_init();
#endif

	FrmGotoForm(calcPrefs.form);
	
	return 0;
}

static void 
StopApplication(void)
{	
	FrmSaveAllForms();
	FrmCloseAllForms();

	/* close all modules */
	result_destroy();
	
	mlib_close_mathlib();
	
	prefs_save_preferences();
	
	db_close();
	
	history_close();
#ifdef GRAPHS_ENABLED	
	grpref_close();
#endif	
}

char *
print_error(CError err)
{
	char *res;

	res = MemPtrNew(MAX_RSCLEN);
	SysStringByIndex(strErrCodes,err,res,MAX_RSCLEN-1);
	return res;
}

void ChangeForm(Int16 formID)
{
	Int16 activefrm = FrmGetActiveFormID();
	
	if (activefrm==formID)
	  return;
	FrmGotoForm(formID);       
}

static WinHandle saved_window;

void
wait_draw(void)
{
	RectangleType rec,frame;
	UInt16 error;

	frame.topLeft.x = 50;
	frame.topLeft.y = 50;
	frame.extent.x = 60;
	frame.extent.y = 20;

	rec.topLeft.x = 48;
	rec.topLeft.y = 48;
	rec.extent.x = 64;
	rec.extent.y = 24;

	saved_window = WinCreateOffscreenWindow(rec.extent.x,rec.extent.y,
						screenFormat,&error);

	/* Save the screen */
	WinCopyRectangle(NULL,saved_window,&rec,0,0,winPaint);
	
	WinEraseRectangle(&rec,2);
	WinDrawRectangleFrame(boldRoundFrame,&frame);
	WinDrawChars("Please wait...",14,55,55);
						  
}

void
wait_erase(void)
{
	RectangleType rec;
		
	rec.topLeft.x = 0;
	rec.topLeft.y = 0;
	rec.extent.x = 64;
	rec.extent.y = 24;

	WinCopyRectangle(saved_window,NULL,&rec,48,48,winPaint);
	WinDeleteWindow(saved_window,0);
}


static Boolean 
ApplicationHandleEvent(EventPtr event)
{
	FormPtr  frm;
	Int16    formId;
	Boolean  handled = false;
	
	if (event->eType == frmLoadEvent) {
		//Load the form resource specified in the event then activate it
		formId = event->data.frmLoad.formID;
		frm = FrmInitForm(formId);
		FrmSetActiveForm(frm);
		// Set the event handler for the form.  The handler of the currently
		// active form is called by FrmDispatchEvent each time it is called
		switch (formId) {
#ifdef GRAPHS_ENABLED			
		case grSetupForm:
			FrmSetEventHandler(frm,GraphSetupHandleEvent);
			break;
		case grPrefForm:
			FrmSetEventHandler(frm,GraphPrefsHandleEvent);
			break;
		case grTableForm:
			FrmSetEventHandler(frm,GraphTableHandleEvent);
			break;
		case frmGraph:
			FrmSetEventHandler(frm,GraphFormHandleEvent);
			break;
#endif /* GRAPHS_ENABLED */
		case prefForm:
			FrmSetEventHandler(frm,PreferencesHandleEvent);
			break;
		case frmInteger:
		case frmBasic:
		case frmBasicS:
		case frmScientific:
			FrmSetEventHandler(frm, MainFormHandleEvent);
			break;
		case defForm:
			FrmSetEventHandler(frm, DefmgrHandleEvent);
			break;
		case finForm:
			FrmSetEventHandler(frm, FinHandleEvent);
			break;
		case varEntryForm:
			FrmSetEventHandler(frm,VarmgrEntryHandleEvent);
			break;
		case memoImportForm:
			FrmSetEventHandler(frm,MemoImportHandleEvent);
			break;
#ifdef SPECFUN_ENABLED
		case slvForm:
			FrmSetEventHandler(frm, SolverHandleEvent);
			break;
		case frmListEdit:
			FrmSetEventHandler(frm,ListEditHandleEvent);
			break;
		case frmMatrix:
			FrmSetEventHandler(frm,MatrixHandleEvent);
			break;
#endif
		}	
		handled = true;	       
	}
	
	return handled;
}

Int32 evtTimeout = evtWaitForever;

void
calc_nil_timeout(Int32 timeout)
{
	evtTimeout = timeout;
}

static void 
EventLoop(void)
{
	EventType  event;
	UInt16      error;
	
	do {
		EvtGetEvent(&event, evtTimeout);
		if (! SysHandleEvent(&event))
		  if (! MenuHandleEvent(0, &event, &error))
		    if (! ApplicationHandleEvent(&event))
		      FrmDispatchEvent(&event);
	} while (event.eType != appStopEvent);
}

UInt32
PilotMain(UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags)
{
	Err err=0;
	
	if (launchCode == sysAppLaunchCmdNormalLaunch) { 
		if ((err = StartApplication()) == 0) {
			EventLoop();
			StopApplication();
		}
	}
	else
	  return sysErrParamErr;
	
	return err;
}
